import 'package:ml_algo/src/common/serializable/serializable.dart';
import 'package:ml_algo/src/retrieval/kd_tree/helpers/create_kd_tree.dart';
import 'package:ml_algo/src/retrieval/kd_tree/helpers/create_kd_tree_from_iterable.dart';
import 'package:ml_algo/src/retrieval/kd_tree/kd_tree_impl.dart';
import 'package:ml_algo/src/retrieval/kd_tree/kd_tree_split_strategy.dart';
import 'package:ml_algo/src/retrieval/neighbour.dart';
import 'package:ml_dataframe/ml_dataframe.dart';
import 'package:ml_linalg/distance.dart';
import 'package:ml_linalg/dtype.dart';
import 'package:ml_linalg/matrix.dart';
import 'package:ml_linalg/vector.dart';

/// KD-tree - an algorithm that provides efficient data retrieval by splitting
/// the whole searching space into partitions in form of binary tree which means
/// that data querying on average will take O(log(n)) time
///
/// One can use this algorithm to perform KNN-search. It's recommended to use
/// [KDTree] when the number of the input data columns is much less than the
/// number of rows of the data - in this case, the search will be more efficient
abstract class KDTree implements Serializable {
  /// [points] Data points which will be used to build the tree.
  ///
  /// [leafSize] A number of points on a leaf node.
  ///
  /// The bigger the number, the less effective search is. If [leafSize] is
  /// equal to the number of [points], a regular KNN-search will take place.
  ///
  /// Small [leafSize] leads to an increasing amount of time for tree building,
  /// but querying will be very fast
  ///
  /// [dtype] A data type which will be used to convert raw data from [points]
  /// into internal numerical representation
  ///
  /// [splitStrategy] Describes how to choose a split dimension. Default value
  /// is [KDTreeSplitStrategy.inOrder]
  ///
  /// if [splitStrategy] is [KDTreeSplitStrategy.inOrder], dimension for data
  /// splits will be chosen one by one in order, in this case, tree building is
  /// very fast
  ///
  /// if [splitStrategy] is [KDTreeSplitStrategy.largestVariance], dimension with
  /// the widest column (in terms of variance) will be chosen to split the data
  ///
  /// [KDTreeSplitStrategy.largestVariance] results in more balanced tree,
  /// but this strategy takes much more time to build the tree than [KDTreeSplitStrategy.inOrder]
  factory KDTree(DataFrame points,
          {int leafSize = 1,
          DType dtype = DType.float32,
          KDTreeSplitStrategy splitStrategy = KDTreeSplitStrategy.inOrder}) =>
      createKDTree(points, leafSize, dtype, splitStrategy);

  /// [pointsSrc] Data points which will be used to build the tree.
  ///
  /// [leafSize] A number of points on a leaf node.
  ///
  /// The bigger the number, the less effective search is. If [leafSize] is
  /// equal to the number of [pointsSrc], a regular KNN-search will take place.
  ///
  /// Extremely small [leafSize] leads to ineffective memory usage since in
  /// this case a lot of kd-tree nodes will be allocated
  ///
  /// [dtype] A data type which will be used to convert raw data from [points]
  /// into internal numerical representation
  ///
  /// [splitStrategy] Describes how to choose a split dimension. Default value
  /// is [KDTreeSplitStrategy.inOrder]
  ///
  /// if [splitStrategy] is [KDTreeSplitStrategy.largestVariance], dimension with
  /// the widest column (in terms of variance) will be chosen to split the data
  ///
  /// if [splitStrategy] is [KDTreeSplitStrategy.inOrder], dimension for data
  /// splits will be chosen one by one in order
  ///
  /// [KDTreeSplitStrategy.largestVariance] provides more accurate KNN-search,
  /// but this strategy takes much more time to build the tree than [KDTreeSplitStrategy.inOrder]
  factory KDTree.fromIterable(Iterable<Iterable<num>> pointsSrc,
          {int leafSize = 1,
          DType dtype = DType.float32,
          KDTreeSplitStrategy splitStrategy = KDTreeSplitStrategy.inOrder}) =>
      createKDTreeFromIterable(pointsSrc, leafSize, dtype, splitStrategy);

  /// Creates a [KDTree] instance from a JSON serializable representation of the
  /// tree. The constructor works in conjunction with [saveAsJson] method.
  ///
  /// Example:
  ///
  /// ```dart
  /// import 'dart:convert';
  /// import 'dart:io';
  /// import 'package:ml_algo/kd_tree.dart';
  ///
  /// void main() async {
  ///   final tree = KDTree.fromIterable([
  ///     [10, 30, 44, 55],
  ///     [11, 70, 41, 75],
  ///     [12, 80, 43, 15],
  ///     [13, 90, 46, 65],
  ///     [14, 11, 47, 95],
  ///   ]);
  ///
  ///   // ...
  ///
  ///   await tree.saveAsJson('path/to/json/file.json');
  ///
  ///   // ...
  ///
  ///   final file = await File('path/to/json/file.json').readAsString();
  ///   final encodedTree = jsonDecode(file) as Map<String, dynamic>;
  ///   final restoredTree = KDTree.fromJson(encodedTree);
  ///
  ///   print(restoredTree);
  /// }
  /// ```
  factory KDTree.fromJson(Map<String, dynamic> json) = KDTreeImpl.fromJson;

  /// Points which were used to build the kd-tree
  Matrix get points;

  /// A number of points on a leaf node.
  ///
  /// The bigger the number, the less effective search is. If [leafSize] is
  /// equal to the number of [points], a regular KNN-search will take place.
  ///
  /// Extremely small [leafSize] leads to ineffective memory usage since in
  /// this case a lot of kd-tree nodes will be allocated
  int get leafSize;

  /// Data type for internal representation of [points]
  DType get dtype;

  /// Returns [k] nearest neighbours for [point]
  ///
  /// The neighbour is represented by an index and the distance between [point]
  /// and the neighbour itself. The index is a zero-based index of a point in
  /// the source [points] matrix. Example:
  ///
  /// ```dart
  /// import 'package:ml_algo/ml_algo.dart';
  /// import 'package:ml_dataframe/ml_dataframe.dart';
  /// import 'package:ml_linalg/vector.dart';
  ///
  /// void main() {
  ///   final data = DataFrame([
  ///     [21, 34, 22, 11],
  ///     [11, 33, 44, 55],
  ///     ...,
  ///   ], headerExists: false);
  ///   final kdTree = KDTree(data);
  ///   final neighbours = kdTree.query(Vector.fromList([1, 2, 3, 4]), 2);
  ///
  ///   print(neighbours[0].index); // let's say, it outputs `3` which means that the nearest neighbour is kdTree.points[3]
  /// }
  /// ```
  Iterable<Neighbour> query(Vector point, int k,
      [Distance distance = Distance.euclidean]);

  /// Returns [k] nearest neighbours for [point], [point] is [Iterable] unlike
  /// [query] method where [point] is [Vector] from ml_linalg library
  ///
  /// One can use the method in the scenario apart from ml_algo ecosystem,
  /// since the method doesn't require dependencies such as ml_linalg
  ///
  /// The neighbour is represented by an index and the distance between [point]
  /// and the neighbour itself. The index is a zero-based index of a point in
  /// the source [points] matrix. Example:
  ///
  /// ```dart
  /// import 'package:ml_algo/kd_tree.dart';
  ///
  /// void main() {
  ///   final data = [
  ///     [21, 34, 22, 11],
  ///     [11, 33, 44, 55],
  ///     ...,
  ///   ];
  ///   final kdTree = KDTree.fromIterable(data);
  ///   final neighbours = kdTree.queryIterable([1, 2, 3, 4], 2);
  ///
  ///   print(neighbours[0].index); // let's say, it outputs `3` which means that the nearest neighbour is kdTree.points[3]
  /// }
  /// ```
  Iterable<Neighbour> queryIterable(Iterable<num> point, int k,
      [Distance distance = Distance.euclidean]);
}
